/*
 * XdccDownloader - DccReceiver.java - Copyright © 2013 David Roden
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.pterodactylus.irc;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.util.concurrent.TimeUnit;

import net.pterodactylus.irc.event.DccDownloadFailed;
import net.pterodactylus.irc.event.DccDownloadFinished;
import net.pterodactylus.irc.event.DccSendReceived;
import net.pterodactylus.xdcc.util.io.BandwidthCountingInputStream;

import com.google.common.eventbus.EventBus;
import com.google.common.io.Closeables;
import com.google.common.util.concurrent.AbstractExecutionThreadService;
import org.apache.log4j.Logger;

/**
 * Service that receives a file offered by a {@link DccSendReceived}.
 *
 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
 */
public class DccReceiver extends AbstractExecutionThreadService {

	/** The logger. */
	private static final Logger logger = Logger.getLogger(DccReceiver.class.getName());

	/** The event bus. */
	private final EventBus eventBus;

	/** The address to connect to. */
	private final InetAddress inetAddress;

	/** The port number to connect to. */
	private final int port;

	/** The name of the file being offered. */
	private final String filename;

	/** The size of the file being offered. */
	private final long size;

	/** The output stream to write the file to. */
	private final OutputStream outputStream;

	/** The number of bytes already written. */
	private long progress;

	/** The bandwidth-measuring input stream. */
	private BandwidthCountingInputStream inputStream;

	/**
	 * Creates a new DCC receiver.
	 *
	 * @param inetAddress
	 * 		The address to connect to
	 * @param port
	 * 		The port number to connect to
	 * @param filename
	 * 		The name of the file being downloaded
	 * @param size
	 * 		The size of the file being downloaded, or {@code -1} if the size is not
	 * 		known
	 * @param outputStream
	 * 		The output stream to write the file to
	 */
	public DccReceiver(EventBus eventBus, InetAddress inetAddress, int port, String filename, long size, OutputStream outputStream) {
		this(eventBus, inetAddress, port, filename, 0, size, outputStream);
	}

	/**
	 * Creates a new DCC receiver.
	 *
	 * @param inetAddress
	 * 		The address to connect to
	 * @param port
	 * 		The port number to connect to
	 * @param filename
	 * 		The name of the file being downloaded
	 * @param startOffset
	 * 		The offset at which the download starts in case of a resume
	 * @param size
	 * 		The size of the file being downloaded, or {@code -1} if the size is not
	 * 		known
	 * @param outputStream
	 * 		The output stream to write the file to
	 */
	public DccReceiver(EventBus eventBus, InetAddress inetAddress, int port, String filename, long startOffset, long size, OutputStream outputStream) {
		this.eventBus = eventBus;
		this.inetAddress = inetAddress;
		this.port = port;
		this.filename = filename;
		this.progress = startOffset;
		this.size = size;
		this.outputStream = outputStream;
	}

	//
	// ACCESSORS
	//

	/**
	 * Returns the name of the file being downloaded. The name is not used by the
	 * DCC receiver, it only serves as a kind of identifier.
	 *
	 * @return The name of the file being downloaded
	 */
	public String filename() {
		return filename;
	}

	/**
	 * Returns the size of the file being downloaded. If the size of the file is
	 * not known, {@code -1} is returned.
	 *
	 * @return The size of the file being downloaded, or {@code -1} if the size is
	 *         not known
	 */
	public long size() {
		return size;
	}

	/**
	 * Returns the number of bytes that have already been downloaded.
	 *
	 * @return The number of bytes that have already been downloaded
	 */
	public long progress() {
		return progress;
	}

	/**
	 * Returns the current rate of the download.
	 *
	 * @return The current rate of the download, in bytes/second
	 */
	public long currentRate() {
		return (inputStream != null) ? inputStream.getCurrentRate() : 0;
	}

	/**
	 * Returns the overall rate of the download.
	 *
	 * @return The overall rate of the download, in bytes/second
	 */
	public long overallRate() {
		return (inputStream != null) ? inputStream.getOverallRate() : 0;
	}

	//
	// ABSTRACTEXECUTIONTHREADSERVICE METHODS
	//

	@Override
	protected void run() throws IOException {
		Socket socket = null;
		try {
			socket = new Socket(inetAddress, port);
			socket.setSoTimeout((int) TimeUnit.MINUTES.toMillis(3));
			InputStream socketInputStream = socket.getInputStream();
			inputStream = new BandwidthCountingInputStream(socketInputStream, 5, TimeUnit.SECONDS);
			byte[] buffer = new byte[65536];
			while (isRunning() && ((size == -1) || (progress < size))) {
				int r = inputStream.read(buffer);
				if (r == -1) {
					/* yay, eof! */
					break;
				}
				outputStream.write(buffer, 0, r);
				progress += r;
			}
			outputStream.flush();
			if ((size == -1) || (progress == size)) {
				eventBus.post(new DccDownloadFinished(this));
			} else {
				eventBus.post(new DccDownloadFailed(this, new IOException("Download aborted.")));
			}
		} catch (IOException ioe1) {
			logger.warn("I/O error while receiving DCC!", ioe1);
			eventBus.post(new DccDownloadFailed(this, ioe1));
		} finally {
			Closeables.close(inputStream, true);
			if (socket != null) {
				socket.close();
			}
		}
	}

}
